"""
Create or update a virtualenv on the Storm workers.  This will be done
automatically upon submit, but this command is provided to help with testing
and debugging.
"""

from __future__ import absolute_import, print_function, unicode_literals

import os
from io import open

from fabric.api import (env, execute, parallel, prefix, put, puts, run, show,
                        sudo)
from fabric.contrib.files import exists
from six import string_types

from .common import (add_config, add_environment, add_name, add_options,
                     add_override_name, add_overwrite_virtualenv,
                     add_pool_size, add_requirements, add_user,
                     resolve_options)
from ..util import (activate_env, die, get_config, get_env_config,
                    get_topology_definition, get_topology_from_file)


@parallel
def _create_or_update_virtualenv(virtualenv_root,
                                 virtualenv_name,
                                 requirements_paths,
                                 virtualenv_flags=None,
                                 overwrite_virtualenv=False,
                                 user='root'):
    with show('output'):
        virtualenv_path = '/'.join((virtualenv_root, virtualenv_name))
        if overwrite_virtualenv:
            puts("Removing virtualenv if it exists in {}".format(virtualenv_root))
            cmd = 'rm -rf {}'.format(virtualenv_path)
            if user == env.user:
                run(cmd, warn_only=True)
            else:
                sudo(cmd, warn_only=True, user=user)
        if not exists(virtualenv_path):
            if virtualenv_flags is None:
                virtualenv_flags = ''
            puts("virtualenv not found in {}, creating one.".format(virtualenv_root))
            run("virtualenv {} {}".format(virtualenv_path, virtualenv_flags))

        if isinstance(requirements_paths, string_types):
            requirements_paths = [requirements_paths]
        for requirements_path in requirements_paths:
            puts("Uploading {} to temporary file.".format(requirements_path))
            temp_req = run("mktemp /tmp/streamparse_requirements-XXXXXXXXX.txt")
            put(requirements_path, temp_req)

            puts("Updating virtualenv: {}".format(virtualenv_name))
            cmd = "source {}".format(os.path.join(virtualenv_path, 'bin/activate'))
            with prefix(cmd):
                # Make sure we're using latest pip so options work as expected
                run("pip install --upgrade 'pip~=9.0'", pty=False)
                run("pip install -r {} --exists-action w --upgrade "
                    "--upgrade-strategy only-if-needed".format(temp_req),
                    pty=False)

            run("rm {}".format(temp_req))


def create_or_update_virtualenvs(env_name, topology_name, options,
                                 virtualenv_name=None, requirements_paths=None,
                                 config_file=None, overwrite_virtualenv=False,
                                 user='root'):
    """Create or update virtualenvs on remote servers.

    Assumes that virtualenv is on the path of the remote server(s).

    :param env_name: the name of the environment in config.json.
    :param topology_name: the name of the topology (and virtualenv).
    :param virtualenv_name: the name that we should use for the virtualenv, even
                          though the topology file has a different name.
    :param requirements_paths: a list of paths to requirements files to use to
                               create virtualenv
    :param overwrite_virtualenv: Force the creation of a fresh virtualenv, even
                                 if it already exists.
    :param user: Who to delete virtualenvs as when using overwrite_virtualenv
    """
    config = get_config()
    topology_name, topology_file = get_topology_definition(topology_name, config_file=config_file)
    topology_class = get_topology_from_file(topology_file)
    env_name, env_config = get_env_config(env_name, config_file=config_file)
    if virtualenv_name is None:
        virtualenv_name = topology_name

    config["virtualenv_specs"] = config["virtualenv_specs"].rstrip("/")

    if requirements_paths is None:
        requirements_paths = [os.path.join(config["virtualenv_specs"],
                                           '{}.txt'.format(topology_name))]

    # Check to ensure streamparse is in at least one requirements file
    found_streamparse = False
    for requirements_path in requirements_paths:
        with open(requirements_path, "r") as fp:
            for line in fp:
                if "streamparse" in line:
                    found_streamparse = True
                    break

    if not found_streamparse:
        die("Could not find streamparse in one of your requirements files.  "
            "Checked {}.  streamparse is required for all topologies."
            .format(requirements_paths))

    # Setup the fabric env dictionary
    storm_options = resolve_options(options, env_config, topology_class, topology_name)
    activate_env(env_name, storm_options, config_file=config_file)

    # Actually create or update virtualenv on worker nodes
    execute(_create_or_update_virtualenv, env.virtualenv_root, virtualenv_name,
            requirements_paths,
            virtualenv_flags=options.get('virtualenv_flags'),
            hosts=env.storm_workers,
            overwrite_virtualenv=overwrite_virtualenv,
            user=user)


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('update_virtualenv',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_config(subparser)
    add_environment(subparser)
    add_overwrite_virtualenv(subparser)
    add_name(subparser)
    add_options(subparser)
    add_override_name(subparser)
    add_pool_size(subparser)
    add_requirements(subparser)
    add_user(subparser)


def main(args):
    """ Create or update a virtualenv on Storm workers. """
    env.pool_size = args.pool_size
    create_or_update_virtualenvs(args.environment, args.name, args.options,
                                 virtualenv_name=args.override_name,
                                 requirements_paths=args.requirements,
                                 config_file=args.config,
                                 overwrite_virtualenv=args.overwrite_virtualenv,
                                 user=args.user)



"""
List the currently running Storm topologies.
"""

from __future__ import absolute_import

from ..util import (get_env_config, get_nimbus_client, print_stats_table,
                    ssh_tunnel)
from ..thrift import TopologySummary
from .common import add_config, add_environment, add_timeout


def _list_topologies(nimbus_client):
    """:returns: A list of running Storm topologies"""
    cluster_summary = nimbus_client.getClusterInfo()
    return cluster_summary.topologies


def list_topologies(env_name, timeout=None, config_file=None):
    """Prints out all running Storm topologies"""
    env_name, env_config = get_env_config(env_name, config_file=config_file)
    # Use ssh tunnel with Nimbus if use_ssh_for_nimbus is unspecified or True
    with ssh_tunnel(env_config) as (host, port):
        nimbus_client = get_nimbus_client(env_config, host=host, port=port,
                                          timeout=timeout)
        topologies = _list_topologies(nimbus_client)
    if not topologies:
        print('No topologies found.')
    else:
        columns = [field for field, default in TopologySummary.default_spec]
        # Find values that are the same for all topologies and list those
        # separately to prevent table from being too wide
        if len(topologies) > 1:
            identical_vals = dict(vars(topologies[0]))
            for topology in topologies:
                for column in columns:
                    if column in identical_vals:
                        cur_val = getattr(topology, column)
                        if cur_val != identical_vals[column]:
                            identical_vals.pop(column)
            if identical_vals:
                for key in identical_vals.keys():
                    columns.remove(key)
                print_stats_table('Values identical for all topologies',
                                  identical_vals,
                                  list(identical_vals.keys()),
                                  'l')

        print_stats_table('Topology-specific values', topologies, columns, 'l')


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('list',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_config(subparser)
    add_environment(subparser)
    add_timeout(subparser)


def main(args):
    """ List the currently running Storm topologies """
    list_topologies(args.environment, timeout=args.timeout, config_file=args.config)



"""
Submit a Storm topology to Nimbus.
"""

from __future__ import absolute_import, print_function, unicode_literals

import importlib
import os
import sys
import time
from itertools import chain

import simplejson as json
from fabric.api import env
from pkg_resources import parse_version
from six import itervalues

from ..dsl.component import JavaComponentSpec
from ..thrift import ShellComponent

from ..util import (activate_env, get_config, get_env_config,
                    get_nimbus_client, get_topology_definition,
                    get_topology_from_file, nimbus_storm_version,
                    set_topology_serializer, ssh_tunnel, warn)
from .common import (add_ackers, add_config, add_debug, add_environment,
                     add_name, add_options, add_override_name,
                     add_overwrite_virtualenv, add_pool_size, add_requirements,
                     add_timeout, add_user, add_wait, add_workers,
                     resolve_options)
from .jar import jar_for_deploy
from .kill import _kill_topology
from .list import _list_topologies
from .update_virtualenv import create_or_update_virtualenvs


THRIFT_CHUNK_SIZE = 307200


def get_user_tasks():
    """Get tasks defined in a user's tasks.py and fabfile.py file which is
    assumed to be in the current working directory.

    :returns: a `tuple` (invoke_tasks, fabric_tasks)
    """
    sys.path.insert(0, os.getcwd())
    try:
        user_invoke = importlib.import_module('tasks')
    except ImportError:
        user_invoke = None
    try:
        user_fabric = importlib.import_module('fabfile')
    except ImportError:
        user_fabric = None
    return user_invoke, user_fabric


def is_safe_to_submit(topology_name, nimbus_client):
    """Is topology not in list of current topologies?"""
    topologies = _list_topologies(nimbus_client)
    safe = not any(topology.name == topology_name for topology in topologies)
    return safe


def _kill_existing_topology(topology_name, force, wait, nimbus_client):
    if force and not is_safe_to_submit(topology_name, nimbus_client):
        print("Killing current \"{}\" topology.".format(topology_name))
        sys.stdout.flush()
        _kill_topology(topology_name, nimbus_client, wait=wait)
        while not is_safe_to_submit(topology_name, nimbus_client):
            print("Waiting for topology {} to quit...".format(topology_name))
            sys.stdout.flush()
            time.sleep(0.5)
        print("Killed.")
        sys.stdout.flush()


def _submit_topology(topology_name, topology_class, remote_jar_path, config,
                     env_config, nimbus_client, options=None):
    if options.get('pystorm.log.path'):
        print("Routing Python logging to {}.".format(options['pystorm.log.path']))
        sys.stdout.flush()

    set_topology_serializer(env_config, config, topology_class)

    # Check if topology name is okay on Storm versions that support that
    if nimbus_storm_version(nimbus_client) >= parse_version('1.1.0'):
        if not nimbus_client.isTopologyNameAllowed(topology_name):
            raise ValueError('Nimbus says {} is an invalid name for a '
                             'Storm topology.'.format(topology_name))

    print("Submitting {} topology to nimbus...".format(topology_name), end='')
    sys.stdout.flush()
    nimbus_client.submitTopology(name=topology_name,
                                 uploadedJarLocation=remote_jar_path,
                                 jsonConf=json.dumps(options),
                                 topology=topology_class.thrift_topology)
    print('done')


def _pre_submit_hooks(topology_name, env_name, env_config, options):
    """Pre-submit hooks for invoke and fabric."""
    user_invoke, user_fabric = get_user_tasks()
    pre_submit_invoke = getattr(user_invoke, "pre_submit", None)
    if callable(pre_submit_invoke):
        pre_submit_invoke(topology_name, env_name, env_config, options)
    pre_submit_fabric = getattr(user_fabric, "pre_submit", None)
    if callable(pre_submit_fabric):
        pre_submit_fabric(topology_name, env_name, env_config, options)


def _post_submit_hooks(topology_name, env_name, env_config, options):
    """Post-submit hooks for invoke and fabric."""
    user_invoke, user_fabric = get_user_tasks()
    post_submit_invoke = getattr(user_invoke, "post_submit", None)
    if callable(post_submit_invoke):
        post_submit_invoke(topology_name, env_name, env_config, options)
    post_submit_fabric = getattr(user_fabric, "post_submit", None)
    if callable(post_submit_fabric):
        post_submit_fabric(topology_name, env_name, env_config, options)


def _upload_jar(nimbus_client, local_path):
    upload_location = nimbus_client.beginFileUpload()
    print("Uploading topology jar {} to assigned location: {}"
          .format(local_path, upload_location))
    total_bytes = os.path.getsize(local_path)
    bytes_uploaded = 0
    with open(local_path, 'rb') as local_jar:
        while True:
            print("Uploaded {}/{} bytes".format(bytes_uploaded, total_bytes),
                  end='\r')
            sys.stdout.flush()
            curr_chunk = local_jar.read(THRIFT_CHUNK_SIZE)
            if not curr_chunk:
                break
            nimbus_client.uploadChunk(upload_location, curr_chunk)
            bytes_uploaded += len(curr_chunk)
        nimbus_client.finishFileUpload(upload_location)
        print("Uploaded {}/{} bytes".format(bytes_uploaded, total_bytes))
        sys.stdout.flush()
    return upload_location


def submit_topology(name=None, env_name=None, options=None, force=False,
                    wait=None, simple_jar=True, override_name=None,
                    requirements_paths=None, local_jar_path=None,
                    remote_jar_path=None, timeout=None, config_file=None,
                    overwrite_virtualenv=False, user='root'):
    """Submit a topology to a remote Storm cluster."""
    config = get_config()
    name, topology_file = get_topology_definition(name, config_file=config_file)
    env_name, env_config = get_env_config(env_name, config_file=config_file)
    topology_class = get_topology_from_file(topology_file)
    if override_name is None:
        override_name = name
    if remote_jar_path and local_jar_path:
        warn('Ignoring local_jar_path because given remote_jar_path')
        local_jar_path = None

    # Check if we need to maintain virtualenv during the process
    use_venv = env_config.get('use_virtualenv', True)

    # Check if user wants to install virtualenv during the process
    install_venv = env_config.get('install_virtualenv', use_venv)

    # Setup the fabric env dictionary
    activate_env(env_name)

    # Handle option conflicts
    options = resolve_options(options, env_config, topology_class,
                              override_name)

    # Run pre_submit actions provided by project
    _pre_submit_hooks(override_name, env_name, env_config, options)

    # If using virtualenv, set it up, and make sure paths are correct in specs
    if use_venv:
        virtualenv_name = env_config.get('virtualenv_name', override_name)
        if install_venv:
            create_or_update_virtualenvs(env_name, name, options,
                                         virtualenv_name=virtualenv_name,
                                         requirements_paths=requirements_paths,
                                         config_file=config_file,
                                         overwrite_virtualenv=overwrite_virtualenv,
                                         user=user)
        streamparse_run_path = '/'.join([env.virtualenv_root, virtualenv_name,
                                         'bin', 'streamparse_run'])
        # Update python paths in bolts
        for thrift_bolt in itervalues(topology_class.thrift_bolts):
            inner_shell = thrift_bolt.bolt_object.shell
            if isinstance(inner_shell, ShellComponent):
                if 'streamparse_run' in inner_shell.execution_command:
                    inner_shell.execution_command = streamparse_run_path
        # Update python paths in spouts
        for thrift_spout in itervalues(topology_class.thrift_spouts):
            inner_shell = thrift_spout.spout_object.shell
            if isinstance(inner_shell, ShellComponent):
                if 'streamparse_run' in inner_shell.execution_command:
                    inner_shell.execution_command = streamparse_run_path

    # In case we're overriding things, let's save the original name
    options['topology.original_name'] = name

    # Set parallelism based on env_name if necessary
    for thrift_component in chain(itervalues(topology_class.thrift_bolts),
                                  itervalues(topology_class.thrift_spouts)):
        par_hint = thrift_component.common.parallelism_hint
        if isinstance(par_hint, dict):
            thrift_component.common.parallelism_hint = par_hint.get(env_name)

    if local_jar_path:
        print('Using prebuilt JAR: {}'.format(local_jar_path))
    elif not remote_jar_path:
        # Check topology for JVM stuff to see if we need to create uber-jar
        if simple_jar:
            simple_jar = not any(isinstance(spec, JavaComponentSpec)
                                 for spec in topology_class.specs)

        # Prepare a JAR that doesn't have Storm dependencies packaged
        local_jar_path = jar_for_deploy(simple_jar=simple_jar)

    if name != override_name:
        print('Deploying "{}" topology with name "{}"...'.format(name,
                                                                 override_name))
    else:
        print('Deploying "{}" topology...'.format(name))
    sys.stdout.flush()
    # Use ssh tunnel with Nimbus if use_ssh_for_nimbus is unspecified or True
    with ssh_tunnel(env_config) as (host, port):
        nimbus_client = get_nimbus_client(env_config, host=host, port=port,
                                          timeout=timeout)
        if remote_jar_path:
            print('Reusing remote JAR on Nimbus server at path: {}'
                  .format(remote_jar_path))
        else:
            remote_jar_path = _upload_jar(nimbus_client, local_jar_path)
        _kill_existing_topology(override_name, force, wait, nimbus_client)
        _submit_topology(override_name, topology_class, remote_jar_path, config,
                         env_config, nimbus_client, options=options)
    _post_submit_hooks(override_name, env_name, env_config, options)


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('submit',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_ackers(subparser)
    add_config(subparser)
    add_debug(subparser)
    add_environment(subparser)
    subparser.add_argument('-f', '--force',
                           action='store_true',
                           help='Force a topology to submit by killing any '
                                'currently running topologies with the same '
                                'name.')
    subparser.add_argument('-j', '--local_jar_path',
                           help='Path to a prebuilt JAR to upload to Nimbus. '
                                'This is useful when you have multiple '
                                'topologies that all run out of the same JAR, '
                                'or you have manually created the JAR.')
    add_name(subparser)
    add_options(subparser)
    add_override_name(subparser)
    add_overwrite_virtualenv(subparser)
    add_pool_size(subparser)
    add_requirements(subparser)
    subparser.add_argument('-R', '--remote_jar_path',
                           help='Path to a prebuilt JAR that already exists on '
                                'your Nimbus server. This is useful when you '
                                'have multiple topologies that all run out of '
                                'the same JAR, and you do not want to upload it'
                                ' multiple times.')
    add_timeout(subparser)
    subparser.add_argument('-u', '--uber_jar',
                           help='Build an Uber-JAR even if you have no Java '
                                'components in your topology.  Useful if you '
                                'are providing your own seriailzer class.',
                           dest='simple_jar', action='store_false')
    add_user(subparser)
    add_wait(subparser)
    add_workers(subparser)


def main(args):
    """ Submit a Storm topology to Nimbus. """
    env.pool_size = args.pool_size
    submit_topology(name=args.name, env_name=args.environment,
                    options=args.options, force=args.force, wait=args.wait,
                    simple_jar=args.simple_jar,
                    override_name=args.override_name,
                    requirements_paths=args.requirements,
                    local_jar_path=args.local_jar_path,
                    remote_jar_path=args.remote_jar_path,
                    timeout=args.timeout,
                    config_file=args.config,
                    overwrite_virtualenv=args.overwrite_virtualenv,
                    user=args.user)



import logging
import unittest

from streamparse.cli import common
from streamparse.dsl import Grouping, Stream, Topology
from streamparse.storm import (Bolt, Component, JavaBolt, JavaSpout, ShellBolt,
                               ShellSpout, Spout)

log = logging.getLogger(__name__)


class WordSpout(Spout):
    outputs = ["word"]


class WordCountBolt(Bolt):
    outputs = ["word", "count"]


class MultiStreamWordCountBolt(Bolt):
    outputs = [Stream(fields=['word', 'count']),
               Stream(fields=['all_word_count'], name='sum')]


class DatabaseDumperBolt(Bolt):
    outputs = []


class CliCommonTests(unittest.TestCase):
    """Tests streamparse/cli/common"""
    maxDiff = 1000

    def test_resolve_options(self):
        """Ensure that CLI > topo settings > config.json"""

        class WordCount(Topology):
            config = {"virtualenv_flags": "-p /path/to/python3"}
            word_spout = WordSpout.spec()
            word_bolt = WordCountBolt.spec(inputs=[word_spout])

        cli_options = {}
        env_config = {
            "user": "username",
            "nimbus": "nimbus.example.com:6627",
            "log": {
                "level": "info"
            },
            "virtualenv_root": "/path/to/virtualenvs",
            "virtualenv_flags": "/path/to/python2",
            "ui.port": 8081,
            "options": {
                "supervisor.worker.timeout.secs": 600,
                "topology.message.timeout.secs" : 60,
                "topology.max.spout.pending" : 500
            }
        }

        options = common.resolve_options(cli_options, env_config,
                                         WordCount, "word_count", local_only=True)

        self.assertEqual(options['supervisor.worker.timeout.secs'], 600)
        self.assertEqual(options['virtualenv_flags'], "-p /path/to/python3")

        cli_options = {"virtualenv_flags": "-p /path/to/python3.6"}
        options = common.resolve_options(cli_options, env_config,
                                         WordCount, "word_count", local_only=True)

        self.assertEqual(options['supervisor.worker.timeout.secs'], 600)
        self.assertEqual(options['virtualenv_flags'], "-p /path/to/python3.6")

